###########################################################################
# Top30 is Copyright (C) 2016-2018 Kyle Robbertze <krobbertze@gmail.com>
#
# Top30 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# Top30 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Top30.  If not, see <http://www.gnu.org/licenses/>.
###########################################################################
"""
Manages the Top30 chart and related functions
"""
from __future__ import unicode_literals

import docx2txt
import mutagen
import os
import youtube_dl

import top30

class Chart:
    """
    A Top 30 chart is what is sent detailing the top 30 songs of the week.
    """
    def __init__(self, chart_file, prefix=""):
        self.prefix = prefix;
        self.songs = parse_docx(chart_file)
        for i in range(0, len(self.songs)):
            self.find_song(i)

    def get(self, position, attribute):
        """
        Returns a property from the song at a position in the chart
        """
        return self.songs[position - 1][attribute]

    def get_prefix(self):
        return self.prefix

    def find_song(self, position):
        """
        Finds a song in the download directory. If it is not found
        it attempts to download it.
        """
        song = self.songs[position]
        artist = sanitize(song['artist'])
        title = sanitize(song['title'])
        song_filename = artist + "-" + title
        for filename in os.listdir(top30.SETTINGS.song_directory()):
            if filename.endswith(song_filename + ".ogg") or song_filename in filename:
                self.songs[position]['path'] = os.path.join(top30.SETTINGS.song_directory(),
                                                       filename)
                return
        print(song['title'] + " by " + song['artist'] + " cannot be found")
        url = input("Enter the youtube URL of the song or file location: ")
        if "http://" in url or "https://" in url:
            download_song(url, song_filename)
            self.songs[position]['path'] = os.path.join(top30.SETTINGS.song_directory(),
                                                   song_filename + ".ogg")
            start = input("Enter the start time (mm:ss): ")
            set_start(self.songs[position]['path'], start)
            return
        url = os.path.expanduser(url)
        if os.path.isfile(url):
            os.symlink(url, os.path.join(top30.SETTINGS.song_directory(),
                                    song_filename + ".ogg"))
        else:
            while not os.path.isfile(url) and \
                  not "http://" in url and not "https://" in url:
                print("Unable to find file")
                print(song['title'] + " by " + song['artist'] + " cannot be found")
                url = input("Enter the youtube URL of the song or file location: ")
                if not "http://" in url and not "https://" in url:
                    url = os.path.expanduser(url)
            if "http://" in url or "https://" in url:
                download_song(url, song_filename)
                start = input("Enter the start time (mm:ss): ")
                set_start(os.path.join(top30.SETTINGS.song_directory(),
                                       song_filename + ".ogg"), start)
            else:
                os.symlink(url, os.path.join(top30.SETTINGS.song_directory(),
                                             song_filename + ".ogg"))
        self.songs[position]['path'] = os.path.join(top30.SETTINGS.song_directory(),
                                                    song_filename + ".ogg")

def sanitize(string):
    """
    Sanitizes a string so that it can be used as part of a filename
    """
    string = string.strip().lower()
    string = string.replace(" ", "_").replace("feat.", "feat").replace("ft.", "ft")
    string = string.replace("/", "").replace(",", "").replace("\u2019", "")
    string = string.replace("'", "").replace("&", "").replace("__", "_")
    return string

def parse_docx(filename):
    """
    Parses a docx at filename
    """
    text = docx2txt.process(filename)
    chart_list = text.split("\n")
    chart_list = [i for i in chart_list if i != '' and i != ' '] # Ignore blank lines
    # Just the chart_list items
    chart_list = chart_list[chart_list.index('1'):chart_list.index('30') + 5]
    chart = []
    for i in range(0, len(chart_list), 5):
        chart.append({"artist": chart_list[i + 1], "title": chart_list[i + 2]})
    return chart

def download_song(url, filename):
    """
    Downloads a song at url and converts it into a useable format
    """
    ydl_opts = {
        'outtmpl': os.path.join(top30.SETTINGS.song_directory(), filename + \
                  ".%(ext)s"),
        'format': 'bestaudio',
        'postprocessors': [{
            'key': 'FFmpegExtractAudio',
            'preferredcodec': 'vorbis',
        }],
    }
    with youtube_dl.YoutubeDL(ydl_opts) as ydl:
        ydl.download([url])

def set_start(filename, start_time):
    """
    Writes the start of the song section to file
    """
    song_meta = mutagen.File(filename, easy=True)
    song_meta[top30.SETTINGS.song_start_tag()] = start_time
    song_meta.save()



###########################################################################
# Top30 is Copyright (C) 2016-2018 Kyle Robbertze <krobbertze@gmail.com>
#
# Top30 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# Top30 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Top30.  If not, see <http://www.gnu.org/licenses/>.
###########################################################################
from PyQt5 import QtCore, QtWidgets, QtGui

from mutagen.id3 import COMM
import mutagen
import top30

class ClipListModel(QtCore.QAbstractTableModel):
    def __init__(self):
        QtCore.QAbstractTableModel.__init__(self, parent=None)
        self.horizontal_header = ["Type", "Filename", "Start"]
        self.data = []

    def supportedDragActions(self):
        return QtCore.Qt.MoveAction

    def flags(self, index):
        flags = QtCore.Qt.ItemIsDropEnabled | QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsSelectable 
        if index.isValid():
            if index.column() == self.columnCount() - 1:
                flags |= QtCore.Qt.ItemIsEditable
            flags |= QtCore.Qt.ItemIsDragEnabled
        return flags

    def rowCount(self, parent=QtCore.QModelIndex()):
        return len(self.data)

    def columnCount(self, parent=QtCore.QModelIndex()):
        return 3

    def data(self, index, role):
        if role == QtCore.Qt.DisplayRole or role == QtCore.Qt.EditRole:
            return self.data[index.row()][index.column()]
        return None

    def setData(self, index, value, role):
        if role != QtCore.Qt.EditRole or not index.isValid() or index.column() != self.columnCount() - 1:
            return False
        if value == "":
            self.data[index.row()][index.column()] = None
            self.data[index.row()][0] = "Voice"
        else:
            self.data[index.row()][0] = "Song"
            if not ":" in value:
                value = "00:" + value
            if value.count(":") > 1:
                return False
            minute = value.split(":")[0]
            second = value.split(":")[1]
            if not minute.isdigit() or not second.isdigit():
                return False
            value = "{0:02d}:{1:0>4.1f}".format(int(minute), float(second))
            self.data[index.row()][index.column()] = value

        filename = self.data[index.row()][1]
        song_meta = mutagen.File(filename)
        tag = top30.SETTINGS.song_start_tag()
        if top30.get_format(filename) == "mp3":
            comment = COMM(3, "eng", "", value)
            song_meta['COMM:eng'] = comment
        else:
            song_meta[tag] = value
        song_meta.save()
        return True;

    def headerData(self, index, orientation, role):
        if role != QtCore.Qt.DisplayRole:
            return None
        if orientation == QtCore.Qt.Horizontal:
            return self.horizontal_header[index]
        else:
            return str(index + 1)

    def appendRow(self, row):
        self.insertRows(self.rowCount(),[row])

    def insertRows(self, row, rows, parent = QtCore.QModelIndex()):
        self.beginInsertRows(parent, row, row + len(rows) - 1)
        for i in range(len(rows)):
            row_i = row + i
            self.data.insert(row_i, rows[i])
        self.endInsertRows()

    def insertRow(self, row, row_data):
        self.insertRows(row, [row_data])

    def removeRows(self, row, count, parent=QtCore.QModelIndex()):
        self.beginRemoveRows(parent, row, row + count - 1)
        for i in range(count):
            del self.data[row + count - 1]
        self.endRemoveRows()
        return True

    def removeRow(self, row):
        self.removeRows(row, 1)

    def moveRows(self, source_parent, source_first, source_last,
            destination_parent, destination):
        # Handle funny crashing bug
        self.beginMoveRows(source_parent, source_first, source_last,
                           destination_parent, destination)
        items = self.data[source_first:source_last + 1]
        self.data = self.data[:source_first] + self.data[source_last + 1:]
        for i in range(len(items)):
            self.data.insert(destination + i, items[i])
        self.endMoveRows()

class ClipListView(QtWidgets.QTableView):
    def __init__(self, parent=None):
        QtWidgets.QTableView.__init__(self, parent=None)
        self.horizontalHeader().setStretchLastSection(True)
        self.resizeColumnsToContents()
        self.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        self.setSelectionMode(QtWidgets.QAbstractItemView.SingleSelection)

        self.setDropIndicatorShown(True)
        self.setAcceptDrops(True)
        self.setDragEnabled(True)

    def dragEnterEvent(self, event):
        if event.mimeData().hasFormat('application/top30clip') or event.mimeData().hasFormat('audio/ogg') or event.mimeData().hasFormat('audio/mp3'):
            event.accept()
        else:
            event.ignore()

    def dragMoveEvent(self, event):
        event.accept()

    def dropEvent(self, event):
        index = self.indexAt(event.pos())
        if not index.isValid() or index.row() == self.source_index.row():
            return
        source = self.source_index.row()
        destination = index.row()
        if destination == source + 1:
            destination = source
            source += 1
        self.model().moveRows(QtCore.QModelIndex(), source, source,
                              QtCore.QModelIndex(), destination)
        event.accept()

    def mousePressEvent(self, event):
        super(ClipListView, self).mousePressEvent(event)
        self.startDrag(event)

    def startDrag(self, event):
        self.source_index = self.indexAt(event.pos())
        if not self.source_index.isValid():
            return

        drag = QtGui.QDrag(self)

        mimeData = QtCore.QMimeData()
        mimeData.setData("application/top30clip", b"")
        drag.setMimeData(mimeData)

        vis = self.source_index.sibling(self.source_index.row() + 1,
                                        self.source_index.column())
        pixmap = QtGui.QPixmap()
        pixmap = self.grab(self.visualRect(vis))
        drag.setPixmap(pixmap)
        result = drag.exec()

def is_number(number):
    try:
        float(number)
        return True
    except ValueError:
        return False



###########################################################################
# Top30 is Copyright (C) 2016-2018 Kyle Robbertze <krobbertze@gmail.com>
#
# Top30 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# Top30 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Top30.  If not, see <http://www.gnu.org/licenses/>.
###########################################################################
"""
Main UI handlers
"""
import os
import sys
import pkg_resources
from PyQt5 import QtCore, QtGui, QtWidgets, uic

import top30
from top30.top_30_creator import Top30Creator
from top30.clip_list import ClipListModel

class MainWindow(QtWidgets.QMainWindow):
    """
    The main window for the GUI version of top30
    """
    def __init__(self, creator):
        super(MainWindow, self).__init__()
        self.creator = creator
        ui_file = pkg_resources.resource_filename(__name__, 'main_window.ui')
        uic.loadUi(ui_file, self)
        self.init_ui()

    def init_ui(self):
        """ Initialises the UI """
        self.btn_add_clip.clicked.connect(self.on_add_clip_clicked)
        self.btn_move_up.clicked.connect(self.on_move_up_clicked)
        self.btn_move_down.clicked.connect(self.on_move_down_clicked)
        self.btn_delete_clip.clicked.connect(self.on_delete_clip_clicked)
        self.btn_create.clicked.connect(self.on_create_clip_clicked)

        self.act_new.triggered.connect(self.on_new_clicked)
        self.act_exit.triggered.connect(QtWidgets.qApp.quit)
        self.act_create_clip.triggered.connect(self.on_create_clip_clicked)
        self.act_add_clip.triggered.connect(self.on_add_clip_clicked)
        self.act_delete_clip.triggered.connect(self.on_delete_clip_clicked)
        self.act_about.triggered.connect(self.on_about_clicked)

        self.load_settings()
        self.init_table()

        self.move(200, 100)
        self.setWindowTitle("Top 30 Creator")
        self.setWindowIcon(QtGui.QIcon(
            pkg_resources.resource_filename(__name__, 'icon.ico')))
        self.show()

    def init_table(self):
        """ Initialises the table implementation """
        self.clip_model = ClipListModel()
        self.clip_view.setModel(self.clip_model)

    def on_new_clicked(self):
        """ Event listener for the new menu item """
        self.init_table()

    def on_add_clip_clicked(self):
        """ Event listener for the add clip button """
        filenames, mime  = QtWidgets.QFileDialog.getOpenFileNames(self,
                "Add clip", top30.SETTINGS.last_open_directory(),
                "Audio(*.mp3 *.ogg);;All Files(*)")
        for filename in filenames:
            self.add_clip(filename)
        if len(filenames) != 0:
            dirname = os.path.dirname(filenames[0])
            top30.SETTINGS.set_last_open_directory(dirname)

    def on_move_up_clicked(self):
        """ Event listener for the move up button """
        item = self.get_selected_clip()
        if not item == None and item.row() > 0:
            self.clip_model.moveRows(QtCore.QModelIndex(), item.row(),
                                     item.row(), QtCore.QModelIndex(),
                                     item.row() - 1)

    def on_move_down_clicked(self):
        """ Event listener for the move down button """
        item = self.get_selected_clip()
        if not item == None and item.row() < self.clip_model.rowCount() - 1:
            self.clip_model.moveRows(QtCore.QModelIndex(), item.row() + 1,
                                     item.row() + 1, QtCore.QModelIndex(),
                                     item.row())

    def on_delete_clip_clicked(self):
        """ Event listener for the delete clip button """
        item = self.get_selected_clip()
        if not item == None:
            self.clip_model.removeRow(item.row())

    def on_create_clip_clicked(self):
        """ Event listener for the Create Clip button """
        if self.clip_model.rowCount() == 0:
            QtWidgets.QMessageBox.warning(self, "No Clips",
                                          "Please add clips to use")
            return

        self.save_settings()
        rundown_name, file_type = self.save_clip()
        if not rundown_name:
            return

        item = self.clip_model.createIndex(-1, 1)
        item = item.sibling(item.row() + 1, item.column())
        while item.isValid():
            clip = item.data()
            clip_type = item.sibling(item.row(), item.column() - 1).data()
            if item.row() == 0:
                rundown = self.creator.get_start(clip)
            elif item.row() == self.clip_model.rowCount() - 1:
                rundown = self.creator.add_end(clip, rundown)
            elif clip_type == "Song":
                rundown = self.creator.add_song(clip, rundown)
            else:
                rundown = self.creator.add_voice(clip, rundown)
            item = item.sibling(item.row() + 1, item.column())
        self.creator.export(rundown_name, "mp3", rundown)
        QtWidgets.QMessageBox.information(self, "Complete",
                                          "Clip " + rundown_name + " created.")
    def on_about_clicked(self):
        """ Displays the about dialog """
        about_text = "top30\nVersion " + top30.VERSION + \
        "\n\ntop30 automatically creates rundowns of a top 30 chart.\n\n" + \
        "This project comes with NO WARRENTY, to the extent permitted by the " + \
        "law. You may redistribute it under the terms of the GNU General " + \
        "Public License version 3\n\nWritten by Kyle Robbertze"
        QtWidgets.QMessageBox.about(self, "top30", about_text)

    def get_selected_clip(self):
        """ Returns the selected clip in the list """
        row = self.clip_view.selectionModel().selectedRows()
        if len(row) == 0:
            return None
        return row[0]

    def add_clip(self, filename):
        """ Adds a clip to the list """
        try:
            time = self.creator.get_start_time(filename)/1000
            time_string = "{0:02.0f}:{1:0>4.1f}".format(time / 60, time % 60)
            row = ["Song", filename, time_string]
        except KeyError:
            row = ["Voice", filename, None]
        selected_clip = self.get_selected_clip()
        if selected_clip == None:
            self.clip_model.appendRow(row)
        else:
            self.clip_model.insertRow(selected_clip.row() + 1, row)
            self.clip_view.selectRow(selected_clip.row() + 1)

    def load_settings(self):
        """ Populates the text fields with the apropriate settings """
        clip_length = str(top30.SETTINGS.song_length()/1000)
        voice_start_overlap = str(top30.SETTINGS.voice_start_overlap()/1000)
        voice_end_overlap = str(top30.SETTINGS.voice_end_overlap()/1000)
        self.txt_song_length.setText(clip_length)
        self.txt_voice_start.setText(voice_start_overlap)
        self.txt_voice_end.setText(voice_end_overlap)

    def save_settings(self):
        """ Updates the settings temporarily with the values of the text fields """
        clip_length = float(self.txt_song_length.text()) * 1000
        top30.SETTINGS.set_song_length(clip_length)

        voice_start_overlap = float(self.txt_voice_start.text()) * 1000
        top30.SETTINGS.set_voice_start_overlap(voice_start_overlap)

        voice_end_overlap = float(self.txt_voice_end.text()) * 1000
        top30.SETTINGS.set_voice_end_overlap(voice_end_overlap)

    def save_clip(self):
        """ Prompts the user for the file path to create the new rundown in """
        filename = QtWidgets.QFileDialog.getSaveFileName(self, "Add clip",
                                                 top30.SETTINGS.last_save_directory(),
                                                 "Audio (*.mp3)")
        top30.SETTINGS.set_save_directory(os.path.dirname(filename[0]))
        return filename

class UserInterface:
    def run(self, creator):
        app = QtWidgets.QApplication(sys.argv)
        ex = MainWindow(creator)
        sys.exit(app.exec_())



###########################################################################
# Top30 is Copyright (C) 2016-2018 Kyle Robbertze <krobbertze@gmail.com>
#
# Top30 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# Top30 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Top30.  If not, see <http://www.gnu.org/licenses/>.
###########################################################################
"""
Runs the Rundown creator
"""
import argparse
import os

from top30.chart import Chart
from top30.handlers import UserInterface
from top30.top_30_creator import Top30Creator
from top30.settings import Settings

VERSION = "1.6.3"

SETTINGS = Settings()

def get_format(filename):
    """ Returns the file type from a filename """
    extension = os.path.splitext(filename)[1]
    if extension.startswith('.'):
        extension = extension[1:]
    return extension

def cli():
    """
    Main function. Runs the command-line program
    """
    parser = argparse.ArgumentParser()
    parser.add_argument("-c", "--current-chart", dest="current_chart",
                        help="chart document to create the rundowns from, " \
                                "required for command line operation")
    parser.add_argument("-p", "--previous-chart", dest="previous_chart",
                        help="the previous chart document, required for " \
                                "command line operation")
    parser.add_argument("-v", "--version", action="store_true",
                        help="prints the version information and exits")
    args = parser.parse_args()
    creator = Top30Creator()
    if args.version:
        print("top30", VERSION)
        print("This project comes with NO WARRENTY, to the extent permitted by the law.")
        print("You may redistribute it under the terms of the GNU General Public License")
        print("version 3; see the file named LICENSE for details.")
        print("\nWritten by Kyle Robbertze")
        return
    if args.previous_chart == None or args.current_chart == None:
        print("Missing chart file arguments")
        parser.print_help()
        exit(120)
    chart = Chart(args.current_chart)
    previous_chart = Chart(args.previous_chart, "last-week_")
    print("Creating 30 - 21 rundown...")
    creator.create_rundown(30, 21, chart)
    print("Creating 20 - 11 rundown...")
    creator.create_rundown(20, 11, chart)
    print("Creating 10 - 2 rundown...")
    creator.create_rundown(10, 2, chart)
    print("Creating last week's 10 - 1 rundown...")
    creator.create_rundown(10, 1, previous_chart)

def gui():
    creator = Top30Creator()
    gui = UserInterface()
    gui.run(creator)



###########################################################################
# Top30 is Copyright (C) 2016-2018 Kyle Robbertze <krobbertze@gmail.com>
#
# Top30 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# Top30 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Top30.  If not, see <http://www.gnu.org/licenses/>.
###########################################################################
"""
Manages the settings that are used while running
"""
from PyQt5.QtCore import QSettings
import os

import top30

class Settings:
    """
    Holds the settings for the top30 module
    """
    def __init__(self):
        self.organisation = 'UCTRadio'
        self.application = 'top30'
        self.config = QSettings(self.organisation, self.application)
        # Set default values if they do not exist
        changed = False
        if (not self.config.contains('song/length')):
            changed = True
            self.config.setValue('song/length', 10000)
        if not self.config.contains('song/start_tag'):
            changed = True
            self.config.setValue('song/start_tag', 'description')
        if not self.config.contains('song/directory'):
            changed = True
            self.config.setValue('song/directory', '~/music/')
        if not self.config.contains('voice/start_overlap'):
            changed = True
            self.config.setValue('voice/start_overlap', 300)
        if not self.config.contains('voice/end_overlap'):
            changed = True
            self.config.setValue('voice/end_overlap', 1400)
        if not self.config.contains('voice/directory'):
            changed = True
            self.config.setValue('voice/directory', '~/music/')
        if not self.config.contains('reference_dbfs'):
            changed = True
            self.config.setValue('reference_dbfs', -20.0)
        if not self.config.contains('last_save_directory'):
            changed = True
            self.config.setValue('last_save_directory', "~/")
        if changed:
            del self.config
            self.config = QSettings(self.organisation, self.application)

    def song_length(self):
        """ Returns the length of the song section """
        return self.config.value('song/length', type=int)

    def set_song_length(self, length):
        """" Saves a new length of the song section """
        self.config.setValue('song/length', length)
        del self.config
        self.config = QSettings(self.organisation, self.application)

    def song_start_tag(self):
        """ Retuns the tag that indicates the start of the song section """
        return self.config.value('song/start_tag', type=str).lower()

    def song_directory(self):
        """ Returns the directory of the songs """
        song_directory = self.config.value('song/directory', type=str)
        if song_directory.startswith("~"):
            return os.path.expanduser(song_directory)
        return song_directory

    def voice_start_overlap(self):
        """ Returns the voice start overlap length """
        return self.config.value('voice/start_overlap', type=int)

    def set_voice_start_overlap(self, length):
        """" Saves a new voice sart overlap length """
        self.config.setValue('voice/start_overlap', length)
        del self.config
        self.config = QSettings(self.organisation, self.application)

    def voice_end_overlap(self):
        """ Returns the voice end overlap length """
        return self.config.value('voice/end_overlap', type=int)

    def set_voice_end_overlap(self, length):
        """" Saves a new voice sart overlap length """
        self.config.setValue('voice/end_overlap', length)
        del self.config
        self.config = QSettings(self.organisation, self.application)

    def voice_directory(self):
        """ Returns the directory of the voice """
        voice_directory = self.config.value('voice/directory', type=str)
        if voice_directory.startswith("~"):
            return os.path.expanduser(voice_directory)
        return voice_directory

    def set_reference_dbfs(self, level):
        """ Saves a new dBFS level to normalise clips to """
        self.config.setValue('reference_dbfs', level)
        del self.config
        self.config = QSettings(self.organisation, self.application)

    def reference_dbfs(self):
        """ Returns the reference dBFS used to normalise source clips """
        return self.config.value("reference_dbfs", type=float)

    def set_last_open_directory(self, directory):
        """ Sets the last open directory """
        self.config.setValue('last_open_directory', directory)
        del self.config
        self.config = QSettings(self.organisation, self.application)

    def last_open_directory(self):
        """
        Returns the last directory used to open a file. If there is none, the
        song directory is returned
        """
        if self.config.value("last_open_directory", type=str) == None:
            set_last_open_directory(self.song_directory())
        last_open_directory = self.config.value('last_open_directory', type=str)
        if last_open_directory.startswith("~"):
            return os.path.expanduser(last_open_directory)
        return last_open_directory

    def set_save_directory(self, directory):
        """ Sets the last save directory """
        self.config.setValue('last_save_directory', directory)
        del self.config
        self.config = QSettings(self.organisation, self.application)

    def last_save_directory(self):
        """ Returns the last directory used to save a file """
        last_save_directory = self.config.value('last_save_directory', type=str)
        if last_save_directory.startswith("~"):
            return os.path.expanduser(last_save_directory)
        return last_save_directory



###########################################################################
# Top30 is Copyright (C) 2016-2017 Kyle Robbertze <krobbertze@gmail.com>
#
# Top30 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# Top30 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Top30.  If not, see <http://www.gnu.org/licenses/>.
###########################################################################
"""
Top30 is a Python 3 module that allows for the creation of rundowns from charts
as sent out by UCT Radio.
"""
from codecs import open
from setuptools import setup, find_packages

setup(
    name='top30',
    version='1.6.3',
    description='Creates rundowns from top30 charts',
    long_description='Top30 automatically creates rundowns of songs for a Top 30 chart show',
    url='https://gitlab.com/paddatrapper/top30',
    author='Kyle Robbertze',
    author_email='krobbertze@gmail.com',
    license='GPL-3',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Other Audience',
        'Topic :: Multimedia :: Sound/Audio',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',

        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    keywords='radio audio rundown',
    packages=find_packages(),
    install_requires=['mutagen', 'pydub', 'docx2txt', 'youtube_dl', 'PyQt5'],
    package_data={
        'top30': [
            'main_window.ui',
            'icon.ico',
         ],
    },
    entry_points={
        'console_scripts': ['rundown-creator=top30:cli'],
        'gui_scripts': ['top30=top30:gui'],
    },
)



###########################################################################
# Top30 is Copyright (C) 2016-2018 Kyle Robbertze <krobbertze@gmail.com>
#
# Top30 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# Top30 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Top30.  If not, see <http://www.gnu.org/licenses/>.
###########################################################################
"""
Creates rundowns from a chart
"""
import os

from mutagen.easyid3 import EasyID3
import mutagen
from pydub import AudioSegment

import top30
import top30.handlers
from top30.chart import Chart

class Top30Creator:
    """
    Class for creating rundowns from a start to stop
    """
    def create_rundown(self, start, end, current_chart):
        """
        Creates a rundown from start to end (both inclusive). If there is a
        directory that is added to the file paths and the intro/outro names.
        """
        voice_dir = top30.SETTINGS.voice_directory()
        prefix = current_chart.get_prefix()

        intro = os.path.join(voice_dir, "{0}{1:02d}-{2:02d}_intro.ogg".format(prefix, int(start), int(end)))
        if not os.path.exists(intro):
            intro = intro[:-3] + "mp3"
        rundown = self.get_start(intro)
        song_file = current_chart.get(start, 'path')
        rundown = self.add_song(song_file, rundown)

        for i in range(start - 1, end - 1, -1):
            voice_file = voice_dir + "/" + str(i) + ".ogg"
            if not os.path.exists(voice_file):
                voice_file = voice_file[:-3] + "mp3"
            rundown = self.add_voice(voice_file, rundown)
            song_file = current_chart.get(i, 'path')
            rundown = self.add_song(song_file, rundown)

        outro = os.path.join(voice_dir, "{0}{1:02d}-{2:02d}_outro.ogg".format(prefix, int(start), int(end)))
        if not os.path.exists(outro):
            outro = outro[:-3] + "mp3"

        rundown = self.add_end(outro, rundown)
        rundown_name = "rundown-{0}{1:02d}-{2:02d}".format(prefix, int(start), int(end))
        self.export(rundown_name, "mp3", rundown)

    def add_voice(self, voice_file, rundown):
        """ Adds a voice segment to the rundown """
        voice = AudioSegment.from_file(voice_file, format=top30.get_format(voice_file))
        voice = self.normalise(voice)
        rundown = rundown.overlay(voice[:top30.SETTINGS.voice_start_overlap()],
                                  position=-top30.SETTINGS.voice_start_overlap())
        return rundown.append(voice[top30.SETTINGS.voice_start_overlap():],
                              crossfade=0.5*1000)

    def add_song(self, song_file, rundown):
        """ Adds a song segment to the rundown """
        start_time = self.get_start_time(song_file)

        song = AudioSegment.from_file(song_file, format=top30.get_format(song_file))
        song = song[start_time:start_time + top30.SETTINGS.song_length()]
        song = self.normalise(song)
        song = song.overlay(rundown[-top30.SETTINGS.voice_end_overlap():])
        return rundown[:-top30.SETTINGS.voice_end_overlap()].append(song, crossfade=0)

    def get_start_time(self, filename):
        """
        Returns the start time of song segment in miliseconds. This is read
        from the file's metadata
        """
        song_meta = mutagen.File(filename, easy=True).tags
        tag = top30.SETTINGS.song_start_tag()
        if top30.get_format(filename) == "mp3":
            song_meta.RegisterTextKey(tag, "COMM::eng")
        try:
            time_code = song_meta[tag][0]
        except KeyError:
            time_code = song_meta['comment'][0]
        if not ":" in time_code:
            raise KeyError("Time code not found")
        time_code_sections = time_code.split(":")
        if len(time_code_sections) != 2:
            raise KeyError("Time code not found")
        try:
            song_length = float(time_code.split(':')[0]) * 60 + \
                          float(time_code.split(':')[1])
            song_length *= 1000
            return song_length
        except:
            raise KeyError("Time code not found")

    def get_start(self, filename):
        """ Returns the intro voice segment """
        return self.normalise(AudioSegment.from_file(filename,
                                                format=top30.get_format(filename)))

    def add_end(self, filename, rundown):
        """ Adds the outro voice segment """
        outro = self.normalise(AudioSegment.from_file(filename,
                                       format=top30.get_format(filename)))
        return rundown.append(outro, crossfade=0)

    def export(self, filename, file_type, rundown):
        """ Exports the rundown as an audio file """
        if top30.get_format(filename) != file_type:
            filename = filename + "." + file_type
        rundown.export(filename, format=file_type)

    def normalise(self, clip):
       required_gain = top30.SETTINGS.reference_dbfs() - clip.dBFS
       return clip.apply_gain(required_gain)
